// ################################################################################
//
//  4x4 Robot Car Release v0.0
//
//  Released:  11/06/2019
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    This code demonstrates how a 4x4 chassis can perform different tasks. It takes
    inputs from an infrared sensor, a keypad, two speed sensors and three ultrasonic distance
    sensors. Refer to the Demo Functions Card for an overview of its controls.

    Key 1 = used for direct and range control.
    Key 2 = used for direct and range control.
    Key 3 = used for direct and range control.
    Key 4 = used for direct and range control.
    Key 5 = used for direct and range control.
    Key 6 = used for direct and range control.
    Key 7 = used for direct and range control.
    Key 8 = used for direct and range control.
    Key 9 = used for direct and range control.
    Key 0 = used for direct and range control.
    Key A = autonomous forward action.
    Key B = backs away from approaching objects.
    Key C = code in motion activities using distance sensors.
    Key D = drives using the keypad 1,2,3,4,5,6,7,8,9 X-Y controls
    Key * = forces a 10 second wait, then resumes drive.
    Key # = soft RESET
    
    This program starts with an LED POST test then waits for user input.

    ALWAYS place the 4x4 car on a stand when testing.
*/
// declare libraries
#include "IRremote.h"

// Declare labels and initialise global variables
#define cmdCtrlMax 20   // max elements in control array
#define echoTrig0 3702  // selected echo trigger range
#define echoTrig1 1092  // selected echo trigger range
#define echoTrig2 1382  // selected echo trigger range
#define echoTrig3 1672  // selected echo trigger range
#define echoTrig4 1962  // selected echo trigger range
#define echoTrig5 2252  // nominal echo trigger range = 30 cm
#define echoTrig6 2542  // selected echo trigger range
#define echoTrig7 2832  // selected echo trigger range
#define echoTrig8 3122  // selected echo trigger range
#define echoTrig9 3412  // selected echo trigger range
#define intPin0 2       // interrupt pin 0
#define intPin1 3       // interrupt pin 1
#define keyCol_0 36     // keypad pin assignment
#define keyCol_1 38     // keypad pin assignment
#define keyCol_2 40     // keypad pin assignment
#define keyCol_3 42     // keypad pin assignment
#define keyRow_0 28     // keypad pin assignment
#define keyRow_1 30     // keypad pin assignment
#define keyRow_2 32     // keypad pin assignment
#define keyRow_3 34     // keypad pin assignment
#define ledGrnPin 26    // GREEN LED pin assignment
#define ledRedPin 22    // RED LED pin assignment
#define ledYelPin 24    // GREEN LED pin assignment
#define pinEchoCnt 6    // Centre ultrasonic Echo pin
#define pinEchoLft 8    // Left ultrasonic Echo pin
#define pinEchoRht 4    // Right ultrasonic Echo pin
#define pinEnA 45       // PWM Enable A - RH Wheels
#define pinEnB 46       // PWM Enable B - LH Wheels
#define pinIn1 47       // PWM In1 - LH Wheels
#define pinIn2 49       // PWM In2 - LH Wheels
#define pinIn3 51       // PWM In3 - RH Wheels
#define pinIn4 53       // PWM In4 - RH Wheels
#define pinSpdLft 18    // Left speed sensor
#define pinSpdRht 19    // Right speed sensor
#define pinTime0 13     // scope time trigger signal 10us pulse
#define pinTime1 12     // scope time signal main loop pulse
#define pinTrigCnt 7    // Centre ultrasonic Trig pin
#define pinTrigLft 9    // Left ultrasonic Trig pin
#define pinTrigRht 5    // Right ultrasonic Trig pin
#define pinVin 0        // voltage reference / 2 input
#define receiver  11    // pin 3 of IR receiver to Arduino digital pin 11

volatile int anyAVal;   // any temp value
volatile int anyBVal;   // any temp value
int autoPWMA;           // baseline speed set for autonomous mode
int autoPWMB;           // baseline speed set for autonomous mode
int autoSTOPPED;        // counts stationary conditions to trigger 'move' response
int cmdCtrlCode[cmdCtrlMax]; // array containing control codes
int cmdCtrlData[cmdCtrlMax]; // array containing control data
int cmdCtrlPnt;         // array pointer 0 - cmdCtrlMax-1
int cmdCtrlVal;         // any value used in commands
int demoCnt;            // a counter used by any demo
int demoCnt1;           // a counter used by any demo
int demoCntMax;         // any max value
int demoFlag;           // a flag value used by any demo
int demoInc;            // a increment used by any demo
int demoPnt;            // pointer used by any demo
int demoPntMin;         // minimum pointer value in any demo
int demoState;          // main state machine for demo functions
int demoStateOld;       // previous task when keypad pressed
int demoTask;           // current demo task
int demoWait;           // count down wait used in any demo
volatile unsigned long distanceCntLft; // left-hand distance count
volatile unsigned long distanceCntRht; // left-hand distance count
int ditherCnt;          // sets the time period for dither = 2 * ditherCnt
int ditherLmt;          // threshold at which dither is switched off
int ditherON;           // 0 = OFF, 1 - 4 tasks
int ditherPWMA;         // dither amplitude applied to PWMA
int ditherPWMB;         // dither amplitude applied to PWMB
int ditherSpd;          // value which sets the dither period
unsigned long durationCnt; // measured time on centre sensor
unsigned long durationLft; // measured time on left sensor
unsigned long durationRht; // measured time on right sensor
unsigned long echoMax;  // max pulse timeout 5 ms
unsigned long echoTrig; // echo pulse duration trigger point. 1 of 10 values
int irInc;              // increment when using IR joypad
int irIncSpCtrl;        // speed control increment when using IR joypad
unsigned long interval; // main loop interval
volatile unsigned long intLftMicros;  // left-hand interupt trigger point 
unsigned long intLftMicrosOld;        // previous left-hand interupt trigger point 
volatile unsigned long intLftPrev;    // previous left-hand interupt period 
volatile unsigned long intRhtMicros;  // right-hand interupt trigger point 
unsigned long intRhtMicrosOld;        // previous left-hand interupt trigger point 
volatile unsigned long intRhtPrev;    // previous right-hand interupt period 
int keyDebounce;
int keyDown;            // = 1 if down, = 0 if up
int keyNew;             //most recent key pressed
int keyState;           // state mode for debounce, etc
int keyVal;             // temporary key pressed value, can be -1
int keyValOld;          // previous key pressed
unsigned long lastMicros; // previous micros() reading
int ledBlinkCtr;        // 0 - 255 blink centre value
int ledBlinkLft;        // 0 - 255 blink left value
int ledBlinkOn;         // > 0 then blink LED's based on ledBlinkLft,...
int ledBlinkRht;        // 0 - 255 blink right value
int ledOnCnt;           // if > 1 then turn LED's ON every main loop cycle
int loopCnt;            // 1-10 counter used for 1 second service routine
unsigned int loopDelay; // main loop delay in microseconds
long periodDef;         // period assigned to stop/default condition
volatile long periodLft; // left-hand interupt period
volatile long periodLftAv; // rolling average of 10 left-hand interupt periods
long periodLftLast;     // anti-lock period check
long periodLftTgt;      // left-hand period target for speed control
long periodMax;         // longest period/slowest speed
long periodMin;         // shortest period/fastest speed
volatile long periodRht;    // right-hand interupt period
volatile long periodRhtAv;  // rolling average of 10 right-hand interupt periods
long periodRhtLast;     // anti-lock period check
long periodRhtTgt;      // right-hand period target for speed control
int rawVin;             // Vin value read from analogue pin 0 - 1023
int serialON = 0;       // set to 1 to enable print responses
int spdCtrlCnt;         // period for speed control in 10ms steps
int spdCtrlON = 0;      // speed control; 0 = OFF, 1 = ON
int spdCtrlPrd;         // perdiod value for speed control counter
int spdLftAbs;          // anti-lock counter
int spdLftDir;          // cpeed control channel direction
int spdLckInc;          // increment used in abs speed lock
int spdLckTrg;          // trigger count for speed lock drive
int spdRhtAbs;          // anti-lock counter
int spdRhtDir;          // cpeed control channel direction
volatile unsigned long speedCntLft; // instantaneous left-hand speed count
unsigned long speedCntLftV;         // left-hand speed count last read
volatile unsigned long speedCntRht; // instantaneou Right-hand speed count
unsigned long speedCntRhtV;         // left-hand speed count last read
int speedInc;           // manual speed increment for PWM
int speedMaxF;          // max forward speed value
int speedMinF;          // min forward value
int speedMaxR;          // max reverse speed value
int speedMinR;          // min reverse speed value
int speedPWMA;          // A channel PWM
int speedPWMB;          // B channel PWM
unsigned long speedTgtLft;  // left-hand speed target
unsigned long speedTgtRht;  // right-hand speed target
int steerBiasA;         // a decaying bias applied to channel A
int steerBiasIncA;      // decaying bias increment applied to channel A
int steerBiasB;         // a decaying bias applied to channel B
int steerBiasIncB;      // decaying bias increment applied to channel B
int steerFlag;          // set true by IR in drive mode to return steering fwd

// Declare objects
IRrecv irrecv(receiver);  // create instance of 'irrecv'
decode_results results;   // create instance of 'decode_results'

void setup() {
  // put your setup code here, to run once:
  setDefaults();
  setSpeedIntOFF();
  pinMode(intPin0, INPUT_PULLUP);
  pinMode(intPin1, INPUT_PULLUP);
  pinMode(keyCol_0, INPUT_PULLUP);
  pinMode(keyCol_1, INPUT_PULLUP);
  pinMode(keyCol_2, INPUT_PULLUP);
  pinMode(keyCol_3, INPUT_PULLUP);
  pinMode(pinEchoLft, INPUT);
  pinMode(pinEchoCnt, INPUT);
  pinMode(pinEchoRht, INPUT);
  pinMode(ledGrnPin, OUTPUT);
  pinMode(ledRedPin, OUTPUT);
  pinMode(ledYelPin, OUTPUT);
  pinMode(pinEnA, OUTPUT);
  digitalWrite(pinEnA, LOW);
  pinMode(pinEnB, OUTPUT);
  digitalWrite(pinEnB, LOW);
  pinMode(pinIn1, OUTPUT);
  pinMode(pinIn2, OUTPUT);
  pinMode(pinIn3, OUTPUT);
  pinMode(pinIn4, OUTPUT);
  pinMode(pinTime0,OUTPUT);
  pinMode(pinTime1,OUTPUT);
  pinMode(pinTrigRht, OUTPUT);
  pinMode(pinTrigCnt, OUTPUT);
  pinMode(pinTrigLft, OUTPUT);
  keypadReset();
  irrecv.enableIRIn(); // Start the infrared receiver
  runPOST();
}

// --------------------------------------------------------------------------------

void loop() {
  // In the main loop we scan the keypad and distance sensors.
  // Sensor readings are handled by the specific demo mode.
  // the loop delay is approx. 10 ms
  loopCnt++; if (loopCnt > 9) {loopCnt = 0;} //every 1/10 second
  switch (loopCnt) {
    case 0: loopSR(); break;  //every 1/10 second print values
    case 5: readVin(); break; // measure supply voltage
  }
  if (ledOnCnt > 0) {
    // used to turn LEDs full ON when max drive is achieved
    ledOnCnt--; writeLED(LOW, LOW, LOW); // turn LEDs ON every cycle
    if (ledOnCnt < 2) {writeLED(HIGH, HIGH, HIGH);} // turn LEDs OFF
  } else if (ledBlinkOn > 0) {ledBlinkWrite();}
  if (ditherON > 0) {
      // dither is active so cycle every 2 loops based on ditherSpd
      // dither frequency can vary from 50Hz down to sub-sonic rates
      ditherCnt--;
      if (ditherCnt < 1) {
        ditherCnt = ditherSpd;
        switch (ditherON) {
          case 1:
            ditherPWMA = speedPWMA; if (speedPWMA < 0) {ditherPWMA = -speedPWMA;}
            ditherPWMA = (speedMaxF - ditherPWMA) / 4;
            if (ditherPWMA < ditherLmt) {ditherPWMA = 0;}
            else if (speedPWMA == 0) {ditherPWMA = 0;}
            setDriveSpeeds (speedPWMA + ditherPWMA, speedPWMB - ditherPWMB);
            ditherON = 2; break;
          case 2:
            ditherPWMB = speedPWMB; if (speedPWMB < 0) {ditherPWMB = -speedPWMB;}
            ditherPWMB = (speedMaxF - ditherPWMB) / 4;
            if (ditherPWMB < ditherLmt) {ditherPWMB = 0;}
            else if (speedPWMB == 0) {ditherPWMB = 0;}
            setDriveSpeeds (speedPWMA - ditherPWMA, speedPWMB + ditherPWMB);
            ditherON = 1; break;
        }
      }
  }
  if (spdCtrlON > 0) {
    // perform speed adjustment tasks based on interupt readings
    spdCtrlCnt--; if (spdCtrlCnt < 0) {autoSpeedCtrl();}
  }
  switch (demoState) {
    // main task scheduler for all 16 tasks
    case 0:
      // default task waiting for 1 key press or hash reset
      keypadCheck();
      if (keyVal == -1) {infraredCheck();}
      if (keyVal == 90) {
        // toggle serial monitor ON/OFF
        if (serialON < 1) {serialON = 1; Serial.println("Serial ON" );}
        else {serialON = 0; Serial.println("Serial OFF" );}
        waitIrRx(); keyVal = -1;
      } break;
    case 1:
      // Undefined task
      demoState = 0;
      break;
    case 2:
      // Undefined task
      demoState = 0;
      break;
    case 3:
      // Undefined task
      demoState = 0;
      break;
    case 4:
      // Undefined task
      demoState = 0;
      break;
    case 5:
      // Undefined task
      demoState = 0;
      break;
    case 6:
      // Undefined task
      demoState = 0;
      break;
    case 7:
      // Undefined task
      demoState = 0;
      break;
    case 8:
      // Undefined task
      demoState = 0;
      break;
    case 9:
      // Undefined task
      demoState = 0;
      break;
    case 10:
      // Undefined task
      demoMode10();
      break;
    case 11:
      // backaway mode, retreats from an approaching object
      demoMode11();
      break;
    case 12:
      // define distances to travel after a 'freeze'
      demoMode12();
      break;
    case 13:
      // drive mode with X-Y keypad
      demoMode13();
      break;
    case 14:
      // Undefined task - normally assigned to 'Freeze' mode
      demoState = 0;
      break;
    case 15:
      // RESET task
      softRESET();
      waitKeyDown();  // ensure no key pressed
      waitIrRx(); // ensure IR Rx has stopped
      break;
  }
  digitalWrite(pinTime1,LOW);  // end of main loop
  // main loop delay, nominally 10ms set by interval variable
  while (micros() - lastMicros < interval) {
    // while waiting average period counts
    if (intLftMicros != intLftMicrosOld) {
      // get average pulse period
      periodLftAv = (periodLftAv + periodLftAv + periodLftAv + periodLft) / 4;
      intLftMicrosOld = intLftMicros;
    }  
    if (intRhtMicros != intRhtMicrosOld) {
      // get average pulse period
      periodRhtAv = (periodRhtAv + periodRhtAv + periodRhtAv + periodRht) / 4;
      intRhtMicrosOld = intRhtMicros;
    }
  } lastMicros = micros();
  // put out a 10us timing pulse
  digitalWrite(pinTime0,HIGH);
  delayMicroseconds(10);
  digitalWrite(pinTime0,LOW);
  digitalWrite(pinTime1,HIGH);  // start of main loop
}

// --------------------------------------------------------------------------------

void autoSpeedCtrl() {
  // performs automatic speed control function based on interrupts  
  spdCtrlCnt = spdCtrlPrd; // reset speed control loop counter
  int zspeedPWMA = 0; int zspeedPWMB = 0;
  if (spdLftDir > 0) {
    // LH is going forward
    if (periodLftLast == periodLft) {
      // lock up so increase drive after so many cycles
      spdLftAbs++;
      if (spdLftAbs >= spdLckTrg) {zspeedPWMA = 255; speedPWMA = 50; Serial.println("ABS-L");}
    }
    else {periodLftLast = periodLft; spdLftAbs = 0;}
    if (periodLftTgt < periodLftAv) {speedPWMA = speedPWMA + 4;}
    else if (periodLftTgt > periodLftAv) {speedPWMA = speedPWMA - 4;}
    if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF;}
    else if (speedPWMA < 0) {speedPWMA = 0;}
  }
  else if (spdLftDir < 0) {
    // LH is going backward
    if (periodLftLast == periodLft) {
      // lock up so increase drive after so many cycles
      spdLftAbs++;
      if (spdLftAbs >= spdLckTrg) {zspeedPWMA = -255; speedPWMA = -50; Serial.println("ABS-L");}
    }
    else {periodLftLast = periodLft; spdLftAbs = 0;}
    if (periodLftTgt < periodLftAv) {speedPWMA = speedPWMA - 4;}
    else if (periodLftTgt > periodLftAv) {speedPWMA = speedPWMA + 4;}
    if (speedPWMA < speedMaxR) {speedPWMA = speedMaxR;}
    else if (speedPWMA > 0) {speedPWMA = 0;}
  }
  if (spdLftDir == 0) {
    // if starting from rest
    if (speedPWMA > 0) {spdLftDir = 1;}
    if (speedPWMA < 0) {spdLftDir = -1;}
  }
  
  if (spdRhtDir > 0) {
    // RH is going forward
    if (periodRhtLast == periodRht) {
      // lock up so increase drive after so many cycles
      spdRhtAbs++;
      if (spdRhtAbs >= spdLckTrg) {zspeedPWMB = 255; speedPWMB = 50; Serial.println("ABS-R");}
    }
    else {periodRhtLast = periodRht; spdRhtAbs = 0;}
    if (periodRhtTgt < periodRhtAv) {speedPWMB = speedPWMB + 4;}
    else if (periodRhtTgt > periodRhtAv) {speedPWMB = speedPWMB - 4;}
    if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF;}
    else if (speedPWMB < 0) {speedPWMB = 0;}
  }
  else if (spdRhtDir < 0) {
    // RH is going backward
    if (periodRhtLast == periodRht) {
      // lock up so increase drive after so many cycles
      spdRhtAbs++;
      if (spdRhtAbs >= spdLckTrg) {zspeedPWMB = -255; speedPWMB = -50; Serial.println("ABS-R");}
    }
    else {periodRhtLast = periodRht; spdRhtAbs = 0;}
    if (periodRhtTgt < periodRhtAv) {speedPWMB = speedPWMB - 4;}
    else if (periodRhtTgt > periodRhtAv) {speedPWMB = speedPWMB + 4;}
    if (speedPWMB < speedMaxR) {speedPWMB = speedMaxR;}
    else if (speedPWMB > 0) {speedPWMB = 0;}
  }
  if (spdRhtDir == 0) {
    // if starting from rest
    if (speedPWMB > 0) {spdRhtDir = 1;}
    if (speedPWMB < 0) {spdRhtDir = -1;}
  }
  if (zspeedPWMA == 0) {zspeedPWMA = speedPWMA;}
  if (zspeedPWMB == 0) {zspeedPWMB = speedPWMB;}
  setDriveSpeeds(zspeedPWMA, zspeedPWMB);
}

// --------------------------------------------------------------------------------

void clrCmdArray() {
  // clear all of the control array elements to 0 and reset pointer
  for (cmdCtrlPnt = 0; cmdCtrlPnt < cmdCtrlMax; cmdCtrlPnt++) {
    cmdCtrlCode[cmdCtrlPnt] = 0;
    cmdCtrlData[cmdCtrlPnt] = 0;
  } cmdCtrlPnt = 0;
}

// --------------------------------------------------------------------------------

void demoMode10() {
  // autonomous 'A' task. User sets drive speed and sensor range.
  // keypad 'A' and 'B' are used to control forward speed
  // keypad '0 - 9' are used to set sensor range
  // keypad '*' acts as a 'freeze' function
  unsigned long zEcho;
  int zEchoTrig = echoTrig;
  int zLL = HIGH;
  int zLR = HIGH;
  int zRunTime = 0;
  int zSpeed = 0;
  float zSpeedMax = 2.0 * speedMaxF; // acts as brake and turning gain
  int zSpeedPWMA = 0;
  int zSpeedPWMB = 0;

  switch(demoTask) {
    case 0:
      // initialise the task
      // set all trigger pins LOW to start with
      if (serialON == 1) {Serial.println("Mode A: 0");}
      digitalWrite(pinTrigRht, LOW);
      digitalWrite(pinTrigCnt, LOW);
      digitalWrite(pinTrigLft, LOW);
      delayMicroseconds(2);
      // turn dither on for this demo mode
      setDitherON();
      autoPWMA = 0; autoPWMB = 0; autoSTOPPED = 0;
      echoTrig = echoTrig5; // mid-range sensor sensitivity
      ledOnCnt = 0;
      ledBlinkOn = 1; // turn ON LED blinking;
      ledBlinkCtr = HIGH; ledBlinkLft = HIGH; ledBlinkRht = HIGH;
      setSpeedIntON(); // turn ON speed sensor counting
      demoTask = 1;
      break;
    case 1:
      // ensure no keypad key is pressed before starting
      if (serialON == 1) {Serial.println("Mode A: 1");}
      waitKeyDown();
      keypadScan(1);
      if (keyDown == 0) {demoTask = 2;}
      break;
    case 2:
      // cycles reading sensor ranges
      // if target in range flash corresponding LED
      // drives wheels in reverse if target <1200 - <600
      keyVal = -1; keypadScan(1);
      if (keyVal == -1) {infraredRead();}
      if (keyVal == 22) {keyVal = 10;}  // substitute forward key
      if (keyVal == 28) {keyVal = 11;}  // substitute reverse key
      switch (keyVal) {
        case -1:
          // no key pressed so just respond to a target
          demoCnt++;
          if (demoCnt > 1) {
            // trigger a new sensor every 20ms
            demoCnt = 0;
            switch (demoPnt) {
              case 0:
                // reset speed adjustment counter
                demoCnt1 = 0; demoInc = 1; // reset speed change increment
                // trigger left sensor
                digitalWrite(pinTrigLft, HIGH); delayMicroseconds(10);
                digitalWrite(pinTrigLft, LOW);
                zEcho = pulseIn(pinEchoLft, HIGH, echoMax);
                if (zEcho == 0) {zEcho = echoMax;}
                durationLft = zEcho;
                loopDelay = loopDelay - zEcho;
                demoPnt = 1; break;
              case 1:
                // trigger centre sensor
                digitalWrite(pinTrigCnt, HIGH); delayMicroseconds(10);
                digitalWrite(pinTrigCnt, LOW);
                zEcho = pulseIn(pinEchoCnt, HIGH, echoMax);
                if (zEcho == 0) {zEcho = echoMax;} // ignore erronious values
                durationCnt = zEcho;
                loopDelay = loopDelay - zEcho;
                demoPnt = 2; break;
              case 2:
                // trigger right sensor
                digitalWrite(pinTrigRht, HIGH); delayMicroseconds(10);
                digitalWrite(pinTrigRht, LOW);
                zEcho = pulseIn(pinEchoRht, HIGH, echoMax);
                if (zEcho == 0) {zEcho = echoMax;}
                durationRht = zEcho;
                loopDelay = loopDelay - zEcho;
                demoPnt = 3; break;
              case 3:
                // now respond to a target
                zSpeedPWMA = autoPWMA; // set baseline forward speeds
                zSpeedPWMB = autoPWMB;
                if (durationCnt <= echoTrig) {
                  // centre target, so half sensitivity
                  zSpeed = durationCnt; // force speed to integer
                  zSpeed = (zSpeed - zEchoTrig)/4; // 512 window = 8.8 cm
                  ledBlinkCtr = LOW;
                  zSpeedPWMA = zSpeedPWMA + zSpeed;
                  zSpeedPWMB = zSpeedPWMB + zSpeed;
                } else {ledBlinkCtr = HIGH;}
                
                if (durationLft <= echoTrig) {
                  // object on left, so back-off R-H side
                  zSpeed = durationLft;
                  zSpeed = (zSpeed - zEchoTrig)/2;
                  ledBlinkLft = LOW;
                  if (zSpeed < speedMaxR) {zSpeed = speedMaxR;}
                  if (zSpeed < steerBiasB) {
                    steerBiasB = zSpeed;
                    steerBiasIncB = -steerBiasB / 16;
                    if (steerBiasIncB < 1) {steerBiasIncB = 1;}
                  }
                } else {ledBlinkLft = HIGH;}
                zSpeedPWMB = zSpeedPWMB + steerBiasB;
                steerBiasB = steerBiasB + steerBiasIncB; // decrease bias over time
                if (steerBiasB > 0) {steerBiasB = 0;}
                
                if (durationRht <= echoTrig) {
                  // object on right, so back of L-H side
                  zSpeed = durationRht;
                  zSpeed = (zSpeed - zEchoTrig)/2;
                  ledBlinkRht = LOW;
                  if (zSpeed < speedMaxR) {zSpeed = speedMaxR;}
                  if (zSpeed < steerBiasA) {
                    steerBiasA = zSpeed;
                    steerBiasIncA = -steerBiasA / 16;
                    if (steerBiasIncA < 1) {steerBiasIncA = 1;}
                  }
                } else {ledBlinkRht = HIGH;}
                zSpeedPWMA = zSpeedPWMA + steerBiasA;
                steerBiasA = steerBiasA + steerBiasIncA; // decrease bias over time
                if (steerBiasA > 0) {steerBiasA = 0;}
                
                // ensure reverse speed has not been exceeded?
                if (zSpeedPWMA < speedMaxR) {zSpeedPWMA = speedMaxR; zLL = LOW;};
                if (zSpeedPWMB < speedMaxR) {zSpeedPWMB = speedMaxR; zLR = LOW;};
                
                // test for stationary condition
                if (autoPWMA < speedMinF) {autoSTOPPED = 0;} // ignore
                if (autoPWMA < speedMinF) {autoSTOPPED = 0;} // ignore
                if (zSpeedPWMA < autoPWMA) {autoSTOPPED++;} // stationary left side
                else {autoSTOPPED--;} //moving left side
                if (zSpeedPWMB < autoPWMB) {autoSTOPPED++;} // stationary right side
                else {autoSTOPPED--;} //moving right side
                if (autoSTOPPED < 0) {autoSTOPPED = 0;} // ensure counter does not go negative
                if (autoSTOPPED > 100) {
                  // stationary for too long so initiate random reverse
                  // and swing action based on nearest side object
                  writeLED(HIGH, LOW, HIGH); // turn ON centre LED
                  if (durationLft < durationRht) {
                    // swing to the left 60 - 240 degrees
                    setDriveSpeeds (speedMaxF, speedMaxR);
                    zRunTime = 300 + (300 * random(3)); delay(zRunTime);
                  }
                  else {
                    // swing to the right 60 - 240 degrees
                    setDriveSpeeds (speedMaxR, speedMaxF);
                    zRunTime = 300 + (300 * random(3)); delay(zRunTime);
                  }
                  writeLED(HIGH, HIGH, HIGH); // turn OFF LEDs
                  autoSTOPPED = 0;
                }
                
                speedPWMA = zSpeedPWMA;
                speedPWMB = zSpeedPWMB;
                setDriveSpeeds (speedPWMA, speedPWMB);
                writeLED(zLL, HIGH, zLR);
                demoPnt = 0;
                break;
              case 4:
                // stop scanning here for 1 second
                demoWait--; if (demoWait <1) {demoPnt = 0;}
                break;
            }
          }
          break;
        case 0:
          // key 0 pressed
          echoTrig = echoTrig0; waitKeyDown();
          break;
        case 1:
          // key 1 pressed
          echoTrig = echoTrig1; waitKeyDown();
          break;
        case 2:
          // key 2 pressed
          echoTrig = echoTrig2; waitKeyDown();
          break;
        case 3:
          // key 3 pressed
          echoTrig = echoTrig3; waitKeyDown();
          break;
        case 4:
          // key 4 pressed
          echoTrig = echoTrig4; waitKeyDown();
          break;
        case 5:
          // key 5 pressed
          echoTrig = echoTrig5; waitKeyDown();
          break;
        case 6:
          // key 6 pressed
          echoTrig = echoTrig6; waitKeyDown();
          break;
        case 7:
          // key 7 pressed
          echoTrig = echoTrig7; waitKeyDown();
          break;
        case 8:
          // key 8 pressed
          echoTrig = echoTrig8; waitKeyDown();
          break;
        case 9:
          // key 9 pressed
          echoTrig = echoTrig9; waitKeyDown();
          break;
        case 10:
          // 'A' key pressed so increase forward speed
          demoFlag = 1;
          speedPWMA = autoPWMA;
          if (speedPWMA < speedMaxF) {
            speedPWMA = speedPWMA + demoInc;
            if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF;}
            if (speedPWMA < speedMinF) {speedPWMA = speedMinF;}
            demoFlag = 0;
          }
          speedPWMB = autoPWMB;
          if (speedPWMB < speedMaxF) {
            speedPWMB = speedPWMB + demoInc;
            if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF; ledOnCnt = 50;}
            if (speedPWMB < speedMinF) {speedPWMB = speedMinF;}
            demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          autoPWMA = speedPWMA; autoPWMB = speedPWMB;
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt1++; if (demoCnt1 >3) {demoCnt1 = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode A: A  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          demoWait = 100; demoPnt = 4; //stop sensor scan
          break;
        case 11:
          // ' B' key pressed so if forward speed is greater than
          // speedMinF reduce it to speedMinF
          demoFlag = 1;
          speedPWMA = autoPWMA;
          if (speedPWMA > speedMinF) {
            speedPWMA = speedPWMA - demoInc;
            if (speedPWMA < speedMinF) {speedPWMA = 0;}
            demoFlag = 0;
          }
          speedPWMB = autoPWMB;
          if (speedPWMB > speedMaxR) {
            speedPWMB = speedPWMB - demoInc;
            if (speedPWMB < speedMinF) {speedPWMB = 0;}
            demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          autoPWMA = speedPWMA; autoPWMB = speedPWMB;
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt1++; if (demoCnt1 >3) {demoCnt1 = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode A: B  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          demoWait = 100; demoPnt = 4; //stop sensor scan
          break;
        case 14:
          // Star * key so freeze for 10 seconds
          setDriveSpeeds (0, 0);
          if (serialON == 1) {Serial.print("Mode A: 14  Freeze!" );}
          waitIrRx(); // to avoid a second freeze event
          freezeWait();
          if (keyDown == 1) {
            // keypress occured during freezeWait so abort/Reset
            waitKeyDown();
            waitIrRx();
            demoState = 15;
          }
          else {setDriveSpeeds (speedPWMA, speedPWMB);}
          break;
        case 15:
          // Hash # key so force to demo Mode 0
          demoState = 15;
          break;
        case 24:
          // IR Left pressed so steer left
          steerBiasA = steerBiasA - 64;
          steerBiasIncA = -steerBiasA / 16;
          if (steerBiasA < speedMaxR) {steerBiasA = speedMaxR;}
          break;
        case 26:
          // IR Right pressed so steer right
          steerBiasB = steerBiasB - 64;
          steerBiasIncB = -steerBiasB / 16;
          if (steerBiasB < speedMaxR) {steerBiasB = speedMaxR;}
          break;
      case 90:
        // toggle serial monitor ON/OFF
        if (serialON < 1) {
          serialON = 1;
          Serial.println("Mode A: 90  Serial ON" );
        } else {
          serialON = 0;
          Serial.println("Mode A: 90  Serial OFF" );
        } waitIrRx();
        break;
      }
    }
  }

// --------------------------------------------------------------------------------

void demoMode11() {
  // backaway 'B' mode tasks. Retreats from an approaching object
  switch(demoTask) {
    case 0:
      // initialise the task
      // set all trigger pins LOW to start with
      if (serialON == 1) {Serial.println("Mode B: 0");}
      digitalWrite(pinTrigRht, LOW);
      digitalWrite(pinTrigCnt, LOW);
      digitalWrite(pinTrigLft, LOW);
      // turn dither on for this demo mode
      setDitherON();
      echoTrig = echoTrig1; // short range sensor sensitivity
      ledOnCnt = 0;
      ledBlinkOn = 1; // turn ON LED blinking;
      ledBlinkCtr = HIGH; ledBlinkLft = HIGH; ledBlinkRht = HIGH;
      setSpeedIntON(); // turn ON speed sensor counting
      demoTask = 1;
      break;
    case 1:
      // ensure no keypad key is pressed before starting
      if (serialON == 1) {Serial.println("Mode B: 1");}
      keypadScan(1);
      if (keyDown == 0) {demoTask = 2;}
      break;
    case 2:
      // cycles reading sensor ranges
      // if target in range flash corresponding LED
      // drives wheels in reverse if target <1200 - <600
      keyVal = -1; keypadScan(1);
      if (keyVal == -1) {infraredRead();}
      switch (keyVal) {
        case -1:
          // no key pressed so just respond to a target
          demoCnt++;
          if (demoCnt > 1) {
            // trigger a new sensor every 20ms
            demoCnt = 0;
            unsigned long zEcho;
            switch (demoPnt) {
              case 0:
                digitalWrite(pinTrigLft, HIGH); delayMicroseconds(10);
                digitalWrite(pinTrigLft, LOW);
                zEcho = pulseIn(pinEchoLft, HIGH, echoMax);
                if (zEcho == 0) {zEcho = echoMax;}
                durationLft = zEcho;
                loopDelay = loopDelay - zEcho;
                demoPnt = 1; break;
              case 1:
                //centre sensor
                digitalWrite(pinTrigCnt, HIGH); delayMicroseconds(10);
                digitalWrite(pinTrigCnt, LOW);
                zEcho = pulseIn(pinEchoCnt, HIGH, echoMax);
                if (zEcho == 0) {zEcho = echoMax;} // ignore erronious values
                durationCnt = zEcho;
                loopDelay = loopDelay - zEcho;
                demoPnt = 2; break;
              case 2:
                digitalWrite(pinTrigRht, HIGH); delayMicroseconds(10);
                digitalWrite(pinTrigRht, LOW);
                zEcho = pulseIn(pinEchoRht, HIGH, echoMax);
                if (zEcho == 0) {zEcho = echoMax;}
                durationRht = zEcho;
                loopDelay = loopDelay - zEcho;
                demoPnt = 3; break;
              case 3:
                // now respond to a target
                int zEchoTrig = echoTrig;
                int zLL = HIGH;
                int zLR = HIGH;
                int zSpeed = 0;
                float zSpeedMax = speedMaxF + 30.0;
                int zSpeedPWMA = 0;
                int zSpeedPWMB = 0;
                if (durationCnt <= echoTrig) {
                  // centre target, so half sensitivity
                  zSpeed = durationCnt; // force speed to integer
                  zSpeed = (zSpeed - zEchoTrig)/4; // 512 window = 8.8 cm
                  ledBlinkCtr = LOW;
                  zSpeedPWMA = zSpeed; zSpeedPWMB = zSpeed;
                } else {ledBlinkCtr = HIGH;}
                if (durationLft <= echoTrig) {
                  // target on L-H side
                  zSpeed = durationLft;
                  zSpeed = (zSpeed - zEchoTrig)/2;
                  ledBlinkLft = LOW;
                  zSpeedPWMA = zSpeedPWMA + zSpeed;
                } else {ledBlinkLft = HIGH;}
                if (durationRht <= echoTrig) {
                  // target on R-H side
                  zSpeed = durationRht;
                  zSpeed = (zSpeed - zEchoTrig)/2;
                  ledBlinkRht = LOW;
                  zSpeedPWMB = zSpeedPWMB + zSpeed;
                } else {ledBlinkRht = HIGH;}
                if (zSpeedPWMA < speedMaxR) {zSpeedPWMA = speedMaxR; zLL = LOW;};
                if (zSpeedPWMA > speedMinR) {zSpeedPWMA = 0;};
                if (zSpeedPWMB < speedMaxR) {zSpeedPWMB = speedMaxR; zLR = LOW;};
                if (zSpeedPWMB > speedMinR) {zSpeedPWMB = 0;};
                speedPWMA = zSpeedPWMA;
                speedPWMB = zSpeedPWMB;
                setDriveSpeeds (speedPWMA, speedPWMB);
                if (zLL == LOW || zLR == LOW) {
                  writeLED(zLL, HIGH, zLR); ledBlinkOn = 0;
                } else {setLedBlinkOn();}
                demoPnt = 0;
            }
          }
          break;
        case 1:
          // key 1 pressed
          echoTrig = echoTrig1; waitKeyDown();
          break;
        case 2:
          // key 1 pressed
          echoTrig = echoTrig2; waitKeyDown();
          break;
        case 3:
          // key 1 pressed
          echoTrig = echoTrig3; waitKeyDown();
          break;
        case 4:
          // key 1 pressed
          echoTrig = echoTrig4; waitKeyDown();
          break;
        case 5:
          // key 1 pressed
          echoTrig = echoTrig5; waitKeyDown();
          break;
        case 6:
          // key 1 pressed
          echoTrig = echoTrig6; waitKeyDown();
          break;
        case 7:
          // key 1 pressed
          echoTrig = echoTrig7; waitKeyDown();
          break;
        case 8:
          // key 1 pressed
          echoTrig = echoTrig8; waitKeyDown();
          break;
        case 9:
          // key 1 pressed
          echoTrig = echoTrig9; waitKeyDown();
          break;
        case 0:
          // key 1 pressed
          echoTrig = echoTrig0; waitKeyDown();
          break;
        case 15:
          // Hash # key so force to demo Mode 0
          demoState = 15;
          break;
      case 90:
        // toggle serial monitor ON/OFF
        if (serialON < 1) {
          serialON = 1;
          Serial.println("Mode B: 90  Serial ON" );
        } else {
          serialON = 0;
          Serial.println("Mode B: 90  Serial OFF" );
        } waitIrRx();
        break;
      }
  }
}

// --------------------------------------------------------------------------------

void demoMode12() {
  // in this 'C' mode the user can specify speeds and travel directions
  // keypad 'A' and 'B' are used to control forward speed
  // enter motions using C + 1-9 keys
  // initiate travel using freeze * key
  loopCnt = 0;  // switch off stats print in this mode
  switch(demoTask) {
    case 0:
      // initialise the task
      // set all trigger pins LOW to start with
      if (serialON == 1) {Serial.println("Mode C: 0");}
      // turn dither on for this demo mode
      setDitherON();
      autoPWMA = 0; autoPWMB = 0; autoSTOPPED = 0;
      ledOnCnt = 0;
      cmdCtrlPnt = 0; // initialise pointer for cmd sequence
      cmdCtrlVal = 0; // initialise data value
      demoTask = 1;
      break;
    case 1:
      // ensure no keypad key is pressed before starting
      if (serialON == 1) {Serial.println("Mode C: 1");}
      waitKeyDown();
      keypadScan(1);
      if (keyDown == 0) {demoTask = 2;}
      break;
    case 2:
      // main demo loop. Scans keypad and responds to keys
      keyVal = -1; keypadScan(1);
      if (keyVal == -1) {infraredRead();}
      switch (keyVal) {
        case -1:
          // no key pressed
          if (keyDown == 0) {demoCnt = 0; demoInc = 1;}
          break;
        case 0:
          // '0' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 0;
             if (serialON == 1) {Serial.println("Mode C: 0  KY cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 0;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 0  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 0; break;
        case 1:
          // '1' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 1;
             if (serialON == 1) {Serial.println("Mode C: 1  NW cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 1;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 1  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 1; break;
        case 2:
          // '2' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 2;
             if (serialON == 1) {Serial.println("Mode C: 2  FW cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 2;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 2  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 2; break;
        case 3:
          // '3' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 3;
             if (serialON == 1) {Serial.println("Mode C: 3  NE cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 3;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 3  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 3; break;
        case 4:
          // '4' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 4;
             if (serialON == 1) {Serial.println("Mode C: 4  NL cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 4;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 4  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 4; break;
        case 5:
          // '5' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 5;
             if (serialON == 1) {Serial.println("Mode C: 5  SP cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 5;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 5  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 5; break;
        case 6:
          // '6' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 6;
             if (serialON == 1) {Serial.println("Mode C: 6  NR cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 6;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 6  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 6; break;
        case 7:
          // '7' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 7;
             if (serialON == 1) {Serial.println("Mode C: 7  SW cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 7;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 7  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 7; break;
        case 8:
          // '8' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 8;
             if (serialON == 1) {Serial.println("Mode C: 8  RV cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 8;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 8  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 8; break;
        case 9:
          // '9' key pressed - so store go forward
          if (keyValOld == 12) {
             // last key was also a 'C' so store cmd in sequence
             cmdCtrlCode[cmdCtrlPnt] = 9;
             if (serialON == 1) {Serial.println("Mode C: 9  SE cmd" );}
          }
          else {
             // part of a numeric sequence
             if (cmdCtrlVal > 0) {cmdCtrlVal = 10 * cmdCtrlVal;}
             cmdCtrlVal = cmdCtrlVal + 9;
             if (cmdCtrlVal > 999) {cmdCtrlVal = 999;}
             cmdCtrlData[cmdCtrlPnt] = cmdCtrlVal; // store new value
             if (serialON == 1) {Serial.print("Mode C: 9  Val = " ); Serial.println(cmdCtrlVal);}
          } 
          waitKeyDown(); waitIrRx();
          keyValOld = 9; break;
        case 10:
          // 'A' key pressed so increase forward speed and store value
          demoFlag = 1;
          cmdCtrlCode[cmdCtrlPnt] = 10;
          speedPWMA = autoPWMA;
          if (speedPWMA < speedMaxF) {
            speedPWMA = speedPWMA + demoInc;
            if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF;}
            if (speedPWMA < speedMinF) {speedPWMA = speedMinF;}
            demoFlag = 0;
          }
          speedPWMB = autoPWMB;
          if (speedPWMB < speedMaxF) {
            speedPWMB = speedPWMB + demoInc;
            if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF; ledOnCnt = 50;}
            if (speedPWMB < speedMinF) {speedPWMB = speedMinF;}
            demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          autoPWMA = speedPWMA; autoPWMB = speedPWMB;
          setDriveSpeeds (speedPWMA, speedPWMB);
          cmdCtrlData[cmdCtrlPnt] = autoPWMA; // common speed value
          demoCnt1++; if (demoCnt1 >3) {demoCnt1 = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode C: A  FW Speed = " ); Serial.println(speedPWMA);}
          keyValOld = 10; break;
        case 11:
          // ' B' key pressed so if forward speed is greater than
          // speedMinF reduce it to speedMinF
          demoFlag = 1;
          cmdCtrlCode[cmdCtrlPnt] = 11;
          speedPWMA = autoPWMA;
          if (speedPWMA > speedMinF) {
            speedPWMA = speedPWMA - demoInc;
            if (speedPWMA < speedMinF) {speedPWMA = 0;}
            demoFlag = 0;
          }
          speedPWMB = autoPWMB;
          if (speedPWMB > speedMaxR) {
            speedPWMB = speedPWMB - demoInc;
            if (speedPWMB < speedMinF) {speedPWMB = 0;}
            demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          autoPWMA = speedPWMA; autoPWMB = speedPWMB;
          setDriveSpeeds (speedPWMA, speedPWMB);
          cmdCtrlData[cmdCtrlPnt] = autoPWMA; // common speed value
          demoCnt1++; if (demoCnt1 >3) {demoCnt1 = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode C: B  FW Speed = " ); Serial.println(speedPWMA);}
          keyValOld = 11; break;
        case 12:
          // 'C' key pressed
          speedPWMA = 0; speedPWMB = 0;
          setDriveSpeeds (0, 0); // stop motors if running
          if (keyValOld == 12) {
            // last key was also a 'C' so terminate cmd sequence
            cmdCtrlCode[cmdCtrlPnt] = 12;
            if (serialON == 1) {Serial.println("Mode C: C  - Terminate" );}
          }
          else {
            // start new code element by incrementing pointer
            // attempts to exceed the data array will overwrite the last
            // value in the array.
            if (cmdCtrlPnt < (cmdCtrlMax - 1)) {
              cmdCtrlPnt++; cmdCtrlVal = 0;
            }
            if (serialON == 1) {Serial.print("Mode C: C  - New Cmd - cmdCtrlPnt = " ); Serial.println(cmdCtrlPnt);}
          }
          waitKeyDown(); waitIrRx();
          keyValOld = 12; break;
        case 14:
          // Star * key so freeze for 10 seconds then run stored cmds
          speedPWMA = 0; speedPWMB = 0;
          setDriveSpeeds (0, 0);
          if (serialON == 1) {Serial.println("Mode C: 14  Freeze!" );}
          freezeWait();
          if (keyDown == 1) {
            // keypress occured during freezeWait so abort/Reset
            waitKeyDown();
            waitIrRx();
            demoState = 15;
          }
          else {demoTask = 3; demoPnt = 0;}
          break;
        case 15:
          // Hash # key so force to demo Mode 0
          demoState = 15;
          break;
        case 90:
          // toggle serial monitor ON/OFF
          if (serialON < 1) {
            serialON = 1;
            Serial.println("Mode D: 90  Serial ON" );
          } else {
            serialON = 0;
            Serial.println("Mode D: 90  Serial OFF" );
          } waitIrRx();
          break;
      } break;
    case 3:
      // demoTask = 3
      // runs the cmd sequence stored in the array
      // if any key is pressed it will abort
      // if sequence completes it will go back to demoTask 2
      switch(demoPnt) {
        case 0:
          // initialise
          if (serialON == 1) {Serial.println("Mode C: Initialise run cmd" );}
          cmdCtrlPnt = 0; demoPnt = 1;
          break;
        case 1:
          // load values from the array
          testKeyDown();
          if (keyDown == 1) {demoTask = 2; break;} // exit run
          switch(cmdCtrlCode[cmdCtrlPnt]) {
            case 0:
              // read 0 from array so stop and wait for freeze or reset
             if (serialON == 1) {Serial.println("Mode C: Run cmd 0" );}
              speedPWMA = 0; speedPWMB = 0;
              setDriveSpeeds (0, 0);
              demoTask = 2; break;
            case 1:
              // go forward NW
              if (serialON == 1) {Serial.println("Mode C: Run cmd 1" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = autoPWMA/2; speedPWMB = autoPWMB;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 2:
              // go forward
              if (serialON == 1) {Serial.println("Mode C: Run cmd 2" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = autoPWMA; speedPWMB = autoPWMB;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 3:
              // go forward NE
              if (serialON == 1) {Serial.println("Mode C: Run cmd 3" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = autoPWMA; speedPWMB = autoPWMB/2;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 4:
              // neutral turn to left
              if (serialON == 1) {Serial.println("Mode C: Run cmd 4" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = -autoPWMA; speedPWMB = autoPWMB;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 5:
              // pause for time
              if (serialON == 1) {Serial.println("Mode C: Run cmd 5" );}
              demoWait = cmdCtrlData[cmdCtrlPnt];
              while (demoWait > 0) {
                delay(100); // wait sub-multiples of 100 ms
                demoWait--;
                testKeyDown();
                if (keyDown == 1) {break;} // exit task
              }
              demoPnt = 2; break;
            case 6:
              // neutral turn to right
              if (serialON == 1) {Serial.println("Mode C: Run cmd 6" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = autoPWMA; speedPWMB = -autoPWMB;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 7:
              // go backwards SW
              if (serialON == 1) {Serial.println("Mode C: Run cmd 1" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = -autoPWMA/2; speedPWMB = -autoPWMB;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 8:
              // go backwards
              if (serialON == 1) {Serial.println("Mode C: Run cmd 2" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = -autoPWMA; speedPWMB = -autoPWMB;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 9:
              // go backwards SE
              if (serialON == 1) {Serial.println("Mode C: Run cmd 1" );}
              distanceCntLft = cmdCtrlData[cmdCtrlPnt];
              distanceCntRht = distanceCntLft;
              attachInterrupt(0, intDistCntDwnRht, CHANGE);
              attachInterrupt(1, intDistCntDwnLft, CHANGE);
              speedPWMA = -autoPWMA; speedPWMB = -autoPWMB/2;
              setDriveSpeeds (speedPWMA, speedPWMB);
              demoPnt = 2; break;
            case 10:
              // A - key set motor speed FWD 
              if (serialON == 1) {Serial.println("Mode C: Run cmd 10" );}
              autoPWMA = cmdCtrlData[cmdCtrlPnt]; autoPWMB = autoPWMA;
              cmdCtrlPnt++; break;
            case 11:
              // B - key set motor speed REV
              if (serialON == 1) {Serial.println("Mode C: Run cmd 11" );}
              autoPWMA = cmdCtrlData[cmdCtrlPnt]; autoPWMB = autoPWMA;
              cmdCtrlPnt++; break;
            case 12:
              // C - key reached the end of the command sequence so loop
              demoPnt = 0; break;
          } break;
        case 2:
          // wait for interrupt run to complete
          testKeyDown();
          if (keyDown == 1) {demoTask = 2; break;} // exit run
          if (serialON == 1) {Serial.println("Mode C: Run cmd - Wait for interrupt end" );}
          if (distanceCntLft < 1 || distanceCntRht < 1) {
            // interrupt count down complete so increment task pointer
            // stop motors too
            speedPWMA = 0; speedPWMB = 0;
            ditherPWMA = 0; ditherPWMB= 0;
            setDriveSpeeds (0, 0);
            delay(100); // allow time for chassis to stop
            if (cmdCtrlPnt < (cmdCtrlMax - 1)) {
              cmdCtrlPnt++;  demoPnt = 1;
            }
            else {
              // run out of commands to do so stop process
              demoTask = 2;
            }
          } break;
      }
      break;
  }
}

// --------------------------------------------------------------------------------

void demoMode13() {
  // drive 'D' mode task. Uses the 9 number keys as a joystick control
  int zVal; // any value
  switch(demoTask) {
    case 0:
      // initialise the task
      if (serialON == 1) {Serial.println("Mode D: 0");}
      speedPWMA = 0; speedPWMB = 0;
      ledBlinkOn = 0; // LED blinking not needed in this mode
      setSpeedIntON(); // turn ON speed sensor counting
      steerFlag = 0; // initial steering bias setting
      steerBiasA = 0; steerBiasB = 0;
      demoTask = 1;
      break;
    case 1:
      // ensure no keypad key is pressed before starting
      if (serialON == 1) {Serial.println("Mode D: 1");}
      keypadScan(1);
      if (keyDown == 0) {demoTask = 2;}
      break;
    case 2:
      // main demo loop. Scans keypad and adjusts motors to suit
      keyVal = -1; keypadScan(1);
      if (keyVal == -1) {infraredRead();}
      switch (keyVal) {
        case -1:
          // no key pressed
          if (keyDown == 0) {demoCnt = 0; demoInc = 1;}
          // is IR steering active?
          if (steerFlag > 0) {
            // yes so remove bias after delay
            steerFlag--;
            if (steerFlag < 1) {
              // set speeds back to normal
              speedPWMA = steerBiasA; speedPWMB = steerBiasB;
              setDriveSpeeds (speedPWMA, speedPWMB);
              steerBiasA = 0; steerBiasB = 0;
              if (spdCtrlON > 0) {if (ditherON > 0) {setDitherOFF();}}
            }
            if (serialON == 1) {Serial.print("Mode D: -1  " ); Serial.print(steerBiasA); Serial.print("  "); Serial.println(steerBiasB);}
          }
          break;
        case 1:
          // Go forward NW
          demoFlag = 1;
          zVal = speedMaxF/2;
          if (speedPWMA < zVal) {
            speedPWMA = speedPWMA + demoInc;
            if (speedPWMA > speedMinR && speedPWMA < speedMinF) {
              speedPWMA = speedMinF;
            } demoFlag = 0;
          }
          if (speedPWMA > zVal) {speedPWMA = speedPWMA - demoInc; demoFlag = 0;}
          if (speedPWMB < (speedMaxF)) {
            speedPWMB = speedPWMB + demoInc + demoInc;
            if (speedPWMB > speedMinR && speedPWMB < speedMinF) {
              speedPWMB = speedMinF;
            } demoFlag = 0;
          }
          if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF;}
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 1  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 2:
          // Go forward
          demoFlag = 1;
          if (speedPWMA < speedMaxF) {
            speedPWMA = speedPWMA + demoInc;
            if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF;}
            if (speedPWMA > speedMinR && speedPWMA < speedMinF) {
              speedPWMA = speedMinF;
            } demoFlag = 0;
          }
          if (speedPWMB < speedMaxF) {
            speedPWMB = speedPWMB + demoInc;
            if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF; ledOnCnt = 50;}
            if (speedPWMB > speedMinR && speedPWMB < speedMinF) {
              speedPWMB = speedMinF;
            } demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 2  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 3:
          // Go forward NE
          demoFlag = 1;
          zVal = speedMaxF/2;
          if (speedPWMA < speedMaxF) {
            speedPWMA = speedPWMA + demoInc + demoInc;
            if (speedPWMA > speedMinR && speedPWMA < speedMinF) {
              speedPWMA = speedMinF;
            } demoFlag = 0;
          }
          if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF;}
          if (speedPWMB < zVal) {
            speedPWMB = speedPWMB + demoInc;
            if (speedPWMB > speedMinR && speedPWMB < speedMinF) {
              speedPWMB = speedMinF;
            } demoFlag = 0;
          }
          if (speedPWMB > zVal) {speedPWMB = speedPWMB - demoInc; demoFlag = 0;}
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 3  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 4:
          // Go Left neutral turn
          demoFlag = 1;
          if (speedPWMA > speedMaxR) {
            speedPWMA = speedPWMA - demoInc;
            if (speedPWMA < speedMaxR) {speedPWMA = speedMaxR;}
            if (speedPWMA < speedMinF && speedPWMA > speedMinR) {
              speedPWMA = speedMinR;
            } demoFlag = 0;
          }
          if (speedPWMB < speedMaxF) {
            speedPWMB = speedPWMB + demoInc;
            if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF;}
            if (speedPWMB > speedMinR && speedPWMB < speedMinF) {
              speedPWMB = speedMinF;
            } demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 4  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 5:
          // STOP!
          if (serialON == 1) {Serial.println("Mode D: 5");}
          speedPWMA = 0; speedPWMB = 0;
          setDriveSpeeds (0, 0); setBrake();
          spdLftDir = 0; spdRhtDir = 0;
          periodLft = periodDef; periodRht = periodDef;
          periodLftAv = periodDef; periodRhtAv = periodDef;
          periodLftTgt = 0; periodRhtTgt = 0;
          break;
        case 6:
          // Go right - neutral turn
          demoFlag = 1;
          if (speedPWMA < speedMaxF) {
            speedPWMA = speedPWMA + demoInc;
            if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF;}
            if (speedPWMA > speedMinR && speedPWMA < speedMinF) {
              speedPWMA = speedMinF;
            } demoFlag = 0;
          }
          if (speedPWMB > speedMaxR) {
            speedPWMB = speedPWMB - demoInc;
            if (speedPWMB < speedMaxR) {speedPWMB = speedMaxR;}
            if (speedPWMB < speedMinF && speedPWMB > speedMinR) {
              speedPWMB = speedMinR;
            } demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 6  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 7:
          // Go reverse SW
          demoFlag = 1;
          zVal = speedMaxR/2;
          if (speedPWMA > zVal) {
            speedPWMA = speedPWMA - demoInc;
            if (speedPWMA > speedMinR && speedPWMA < speedMinF) {
              speedPWMA = speedMinR;
            } demoFlag = 0;
          }
          if (speedPWMA < zVal) {speedPWMA = speedPWMA + demoInc; demoFlag = 0;}
          if (speedPWMB > (speedMaxR)) {
            speedPWMB = speedPWMB - demoInc - demoInc;
            if (speedPWMB > speedMinR && speedPWMB < speedMinF) {
              speedPWMB = speedMinR;
            } demoFlag = 0;
          }
          if (speedPWMB < speedMaxR) {speedPWMB = speedMaxR;}
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 7  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 8:
          // Go in reverse
          demoFlag = 1;
          if (speedPWMA > speedMaxR) {
            speedPWMA = speedPWMA - demoInc;
            if (speedPWMA < speedMaxR) {speedPWMA = speedMaxR;}
            if (speedPWMA < speedMinF && speedPWMA > speedMinR) {
              speedPWMA = speedMinR;
            } demoFlag = 0;
          }
          if (speedPWMB > speedMaxR) {
            speedPWMB = speedPWMB - demoInc;
            if (speedPWMB < speedMaxR) {speedPWMB = speedMaxR;}
            if (speedPWMB < speedMinF && speedPWMB > speedMinR) {
              speedPWMB = speedMinR;
            } demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 8  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 9:
          // Go forward NE
          demoFlag = 1;
          zVal = speedMaxR/2;
          if (speedPWMA > speedMaxR) {
            speedPWMA = speedPWMA - demoInc - demoInc;
            if (speedPWMA > speedMinR && speedPWMA < speedMinF) {
              speedPWMA = speedMinR;
            } demoFlag = 0;
          }
          if (speedPWMA < speedMaxR) {speedPWMA = speedMaxR;}
          if (speedPWMB > zVal) {
            speedPWMB = speedPWMB - demoInc;
            if (speedPWMB > speedMinR && speedPWMB < speedMinF) {
              speedPWMB = speedMinR;
            } demoFlag = 0;
          }
          if (speedPWMB < zVal) {speedPWMB = speedPWMB + demoInc; demoFlag = 0;}
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          demoCnt++; if (demoCnt >3) {demoCnt = 0; if (demoInc < 10) {demoInc++;}}
          if (serialON == 1) {Serial.print("Mode D: 9  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 10:
          // A - key, so toggle dither mode ON or OFF
          if (ditherON > 0) {
            // dither is ON so switch it OFF
            setDitherOFF();
            setDriveSpeeds (speedPWMA, speedPWMB); // remove any offset
            if (serialON == 1) {Serial.print("Mode D: 10  " ); Serial.println("Dither OFF");}
          } else {
            // dither is OFF so switch it ON
            setDitherON();
            if (serialON == 1) {Serial.print("Mode D: 10  " ); Serial.println("Dither ON");}
          }
          waitKeyDown(); // wait for key release so doesn't toggle
          waitIrRx();
          break;
        case 12:
          // C - key so toggle speed control ON or OFF
          if (spdCtrlON > 0) {
            // speed control is ON so switch it OFF
            setSpeedOFF();
            if (serialON == 1) {Serial.print("Mode D: 12  " ); Serial.println("Speed Ctrl OFF");}
         } else {
            // speed control is OFF so switch it ON
            setSpeedON();
            if (serialON == 1) {Serial.print("Mode D: 12  " ); Serial.println("Speed Ctrl ON");}
          }
          waitIrRx();
          break;
        case 13:
          // D - key so increasee period of dither, slowing its frequency
          setSpeedOFF(); //disable speed control when dither is active
          if (ditherON < 1) {
            // dither is OFF so switch it ON
            setDitherON();
          } ditherSpd++;
          waitKeyDown();
          break;
        case 14:
          // Star * key so freeze for 10 seconds
          setDriveSpeeds (0, 0);
          if (serialON == 1) {Serial.print("Mode D: 14  Freeze!" );}
          freezeWait();
          if (keyDown == 1) {
            // keypress occured during freezeWait so abort/Reset
            waitKeyDown();
            waitIrRx();
            demoState = 15;
          }
          else {setDriveSpeeds (speedPWMA, speedPWMB);}
          break;
        case 15:
          // Hash # key so force to demo Mode 0 and reset
          demoState = 15;
          break;
        case 22:
          // IR Go forward faster
          demoFlag = 1;
          if (spdCtrlON > 0) {
            if (ditherON > 0) {setDitherOFF();}
            if (spdLftDir == 0) {spdLftDir = 1; periodLftTgt = periodMax;}
            if (spdLftDir > 0) {
              // LH going forward so speed up
              if (periodLftTgt > periodMin) {
                periodLftTgt = periodLftTgt - irIncSpCtrl; demoFlag = 0;}
              if (periodLftTgt < periodMin) {periodLftTgt = periodMin;}
            }
            else {
              // LH going backwards so slow down/reverse direction
              if (periodLftTgt <= periodMax) {
                periodLftTgt = periodLftTgt + irIncSpCtrl + irIncSpCtrl; demoFlag = 0;
              }
              if (periodLftTgt > periodMax) {
                spdLftDir = 1; periodLftTgt = periodMax;}
            }
            if (spdRhtDir == 0) {spdRhtDir = 1; periodRhtTgt = periodMax;}
            if (spdRhtDir > 0) {
              // RH going forward so speed up
              if (periodRhtTgt > periodMin) {
                periodRhtTgt = periodRhtTgt - irIncSpCtrl; demoFlag = 0;}
              if (periodRhtTgt < periodMin) {periodRhtTgt = periodMin;}
            }
            else {
              // RH going backwards so slow down/reverse direction
              if (periodRhtTgt <= periodMax) {
                periodRhtTgt = periodRhtTgt + irIncSpCtrl + irIncSpCtrl; demoFlag = 0;
              }
              if (periodRhtTgt > periodMax) {
                spdRhtDir = 1; periodRhtTgt = periodMax;}
            }
          } else {
            if (ditherON <1) {setDitherON();}
            if (speedPWMA < speedMaxF) {
              speedPWMA = speedPWMA + irInc;
              if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF; ledOnCnt = 50;}
              if (speedPWMA > speedMinR && speedPWMA < speedMinF) {
                speedPWMA = speedMinF;
              } demoFlag = 0;
            }
            if (speedPWMB < speedMaxF) {
              speedPWMB = speedPWMB + irInc;
              if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF; ledOnCnt = 50;}
              if (speedPWMB > speedMinR && speedPWMB < speedMinF) {
                speedPWMB = speedMinF;
              } demoFlag = 0;
            }
            setDriveSpeeds (speedPWMA, speedPWMB);
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          Serial.print("Mode D: 22  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);
          break;
        case 24:
          // IR Go Left neutral turn faster
          demoFlag = 1;
          if (ditherON <1) {setDitherON();}
          if (steerFlag < 1) {
            steerBiasA = speedPWMA; steerBiasB = speedPWMB;
          } steerFlag = 10;
          if (speedPWMA > speedMaxR) {
            speedPWMA = speedPWMA - irInc - irInc;
            if (speedPWMA < speedMaxR) {speedPWMA = speedMaxR;}
            demoFlag = 0;
          } 
          if (speedPWMB < speedMaxF) {
            speedPWMB = speedPWMB + irInc + irInc;
            if (speedPWMB > speedMaxF) {speedPWMB = speedMaxF;}
            demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          if (serialON == 1) {Serial.print("Mode D: 24  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 26:
          // IR Go right - neutral turn faster
          demoFlag = 1;
          if (ditherON <1) {setDitherON();}
          if (steerFlag < 1) {
            steerBiasA = speedPWMA; steerBiasB = speedPWMB;
          } steerFlag = 10;
          if (speedPWMA < speedMaxF) {
            speedPWMA = speedPWMA + irInc + irInc;
            if (speedPWMA > speedMaxF) {speedPWMA = speedMaxF;}
            demoFlag = 0;
          }
          if (speedPWMB > speedMaxR) {
            speedPWMB = speedPWMB - irInc - irInc;
            if (speedPWMB < speedMaxR) {speedPWMB = speedMaxR;}
            demoFlag = 0;
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          setDriveSpeeds (speedPWMA, speedPWMB);
          if (serialON == 1) {Serial.print("Mode D: 26  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);}
          break;
        case 28:
          // IR Go in reverse faster
          demoFlag = 1;
          if (spdCtrlON > 0) {
            if (ditherON > 0) {setDitherOFF();}
            if (spdLftDir ==0 ) {spdLftDir = -1; periodLftTgt = periodMax;}
            if (spdLftDir < 0) {
              // LH going backward so speed up
              if (periodLftTgt > periodMin) {
                periodLftTgt = periodLftTgt - irIncSpCtrl; demoFlag = 0;}
              if (periodLftTgt < periodMin) {periodLftTgt = periodMin;}
            }
            else {
              // LH going forwards so slow down/reverse direction
              if (periodLftTgt <= periodMax) {
                periodLftTgt = periodLftTgt + irIncSpCtrl + irIncSpCtrl; demoFlag = 0;
              }
              if (periodLftTgt > periodMax) {
                spdLftDir = -1; periodLftTgt = periodMax;}
            }
            if (spdRhtDir == 0) {spdRhtDir = -1; periodRhtTgt = periodMax;}
            if (spdRhtDir < 0) {
              // RH going backward so speed up
              if (periodRhtTgt > periodMin) {
                periodRhtTgt = periodRhtTgt - irIncSpCtrl; demoFlag = 0;}
              if (periodRhtTgt < periodMin) {periodRhtTgt = periodMin;}
            }
            else {
              // RH going forwards so slow down/reverse direction
              if (periodRhtTgt <= periodMax) {
                periodRhtTgt = periodRhtTgt + irIncSpCtrl + irIncSpCtrl; demoFlag = 0;
              }
              if (periodRhtTgt > periodMax) {
                spdRhtDir = -1; periodRhtTgt > periodMax;}
            }
          } else {
            if (ditherON <1) {setDitherON();}
            if (speedPWMA > speedMaxR) {
              speedPWMA = speedPWMA - irInc;
              if (speedPWMA < speedMaxR) {speedPWMA = speedMaxR; ledOnCnt = 50;}
              if (speedPWMA < speedMinF && speedPWMA > speedMinR) {
                speedPWMA = speedMinR;
              } demoFlag = 0;
            }
            if (speedPWMB > speedMaxR) {
              speedPWMB = speedPWMB - irInc;
              if (speedPWMB < speedMaxR) {speedPWMB = speedMaxR; ledOnCnt = 50;}
              if (speedPWMB < speedMinF && speedPWMB > speedMinR) {
                speedPWMB = speedMinR;
              } demoFlag = 0;
            }  
            setDriveSpeeds (speedPWMA, speedPWMB);
          }
          if (demoFlag > 0) {ledOnCnt = 100;}
          Serial.print("Mode D: 28  " ); Serial.print(speedPWMA); Serial.print("  "); Serial.println(speedPWMB);
          break;
      case 90:
        // toggle serial monitor ON/OFF
        if (serialON < 1) {
          serialON = 1;
          Serial.println("Mode D: 90  Serial ON" );
        } else {
          serialON = 0;
          Serial.println("Mode D: 90  Serial OFF" );
        } waitIrRx();
        break;
    }
  }
}

// --------------------------------------------------------------------------------

void freezeWait() {
  // wait for key release, then wait 10 seconds
  // if a key is pressed during that time return it
  writeLED(LOW, LOW, LOW);  // turn ON all LEDs
  waitKeyDown();
  waitIrRx();
  // all keypad and IR keys released so now enter freeze wait
  // blink OFF every 1 second interval
  int zFc = 1000; // set delay 1000 * 10mS = 10 seconds
  int zBlink = 100;
  while (zFc > 0) {
    delay(10); zFc--;
    keypadTest();
    if (keyDown == 1) {zFc = -1;} // force an exit
    infraredTest();
    if (keyDown == 1) {zFc = -1;} // force an exit
    zBlink--;
    if (zBlink < 1) {zBlink = 100;}
    if (zBlink < 5) {
      writeLED(HIGH, HIGH, HIGH);  // turn OFF all LEDs 
    }
    else {writeLED(LOW, LOW, LOW);}
  } writeLED(HIGH, HIGH, HIGH); // turn LEDs OFF on exit
  if (zFc == -1) {keyDown = 1;} // flag forced exit
}

// --------------------------------------------------------------------------------

void infraredCheck() {
  // similar to keypadCheck() call in main loop
  // keyVal limited to 0 - 15 to match 4x4 keypad assignments
  infraredRead();
  if (keyVal > -1 && keyVal < 16) {
    writeLED(LOW,LOW,LOW); // turn ON LED's
    waitIrRx(); // wait for no IR codes
    writeLED(HIGH,HIGH,HIGH); // turn ON LED's
    // an IR code has been received
    demoStateOld = demoState;
    demoState = keyVal;
    demoTask = 0;  // re initialise demo task
    demoCnt = 0;  // clear counter
    demoInc = 1;  // set default increment
    demoPnt = 0;  // clear pointer
    setDitherOFF();
    keyValOld = keyVal;
  }
}

// --------------------------------------------------------------------------------

void infraredRead() {
  // check to see if a code has been recieved. If so decode it.
  if (irrecv.decode(&results)) {
    // we have received an IR signal
    translateIR();
    Serial.print("infrared Read = ");
    Serial.println(keyVal);
    irrecv.resume(); // ready to receive the next value
    keyDown = 1;
  }
  else {
    // no IR code recieved
    keyDown = 0;
  }
}

// --------------------------------------------------------------------------------

void infraredTest() {
  // check to see if a code has been recieved?
  if (irrecv.decode(&results)) {
    // we have received an IR signal
    irrecv.resume(); // ready to receive the next value
    keyDown = 1;
  }
  else {
    // no IR code recieved
    keyDown = 0;
  }
}

// --------------------------------------------------------------------------------

void intDistCntDwnLft() {
  // ISR for int.1 left-hand speed sensor
  // once count reaches 0 turn ON brakes and disable interrupt
  distanceCntLft--; //delayMicroseconds(3000); // debounce
  if (distanceCntLft < 1) {
    setBrake();
    Serial.println("L-H Count");
    detachInterrupt(1);
  }
}

// --------------------------------------------------------------------------------

void intDistCntDwnRht() {
  // ISR for int.0 right-hand speed sensor
  // once count reaches 0 turn ON brakes
  distanceCntRht--; //delayMicroseconds(3000); // debounce
  if (distanceCntRht < 1) {
    setBrake();
    Serial.println("R-H Count");
    detachInterrupt(0);
  }
}

// --------------------------------------------------------------------------------

void intSpeedCntLft() {
  // ISR for int.1 left-hand speed counter
  unsigned long zintMicros = micros(); anyAVal++;
  if ((zintMicros - intLftMicros) > 1000) {
    // edge trigger debounced by 1ms
    periodLft = zintMicros - intLftMicros;
    intLftPrev = intLftMicros;
    intLftMicros = zintMicros; speedCntLft++;
  }
}

// --------------------------------------------------------------------------------

void intSpeedCntRht() {
  // ISR for int.0 right-hand speed counter
  unsigned long zintMicros = micros(); anyBVal++;
  if ((zintMicros - intRhtMicros) > 1000) {
    // edge trigger debounced by 1ms
    periodRht = zintMicros - intRhtMicros;
    intRhtPrev = intRhtMicros;
    intRhtMicros = zintMicros; speedCntRht++;
  }
}

// --------------------------------------------------------------------------------

void keypadCheck() {
  // test for a key beiong pressed and if so loop here until released
  keyVal = -1; keypadScan(0);
  while (keyState > 0) {
    // in the process of reading a key
    keypadScan(0);
  }
}

// --------------------------------------------------------------------------------

void keypadLOW() {
  // sets the row drivers LOW so that any column key press
  // will be detected and decoded.
  pinMode(keyRow_0, OUTPUT);
  digitalWrite(keyRow_0,LOW);
  pinMode(keyRow_1, OUTPUT);
  digitalWrite(keyRow_1,LOW);
  pinMode(keyRow_2, OUTPUT);
  digitalWrite(keyRow_2,LOW);
  pinMode(keyRow_3, OUTPUT);
  digitalWrite(keyRow_3,LOW);
}

// --------------------------------------------------------------------------------

void keypadReset() {
  // sets the row drivers LOW and resets counter
  keypadLOW();
  keyDebounce = 0;
  keyState = 0;
}

// --------------------------------------------------------------------------------

void keypadScan(int zMode) {
  // main keypad scan routine
  // zMode = 0  if key pressed change state
  // zMode = 1  just return keyVal don't wait for release
  switch (keyState) {
    case 0:
      // constantly checks for one or more switch presses
      if (digitalRead(keyCol_0) == LOW || digitalRead(keyCol_1) == LOW || digitalRead(keyCol_2) == LOW || digitalRead(keyCol_3) == LOW) {
        // one or more keys have been pressed
        keyState = 1;
        keyDebounce = 10;  // allow 100mS debounce
      }
      else {keyDown = 0;}
      break;
    case 1:
      // keypress occured so wait for end of debounce period
      if (digitalRead(keyCol_0) == LOW || digitalRead(keyCol_1) == LOW || digitalRead(keyCol_2) == LOW || digitalRead(keyCol_3) == LOW) {
        // still low so debounce timeout
        --keyDebounce;
        if (keyDebounce <= 0) {
          // keydown considered valid
          keyDown = 1; keyState = 2;
        }
      }
      else {
        // key has bounced so reset state to 0
        keyState = 0;
      }
      break;
    case 2:
      // debounce over so now decode the key
      //digitalWrite(buzzDrive,HIGH);  // sound buzzer
      keyValOld = keyNew;
      pinMode(keyRow_1, INPUT);
      pinMode(keyRow_2, INPUT);
      pinMode(keyRow_3, INPUT);
      pinMode(keyRow_0, OUTPUT);
      digitalWrite(keyRow_0,LOW);
      delay(1);
      if (digitalRead(keyCol_0) == LOW) { keyVal = 1;}
      else if (digitalRead(keyCol_1) == LOW) { keyVal = 2;}
      else if (digitalRead(keyCol_2) == LOW) { keyVal = 3;}
      else if (digitalRead(keyCol_3) == LOW) { keyVal = 10;} // key 'A'
      else {
        pinMode(keyRow_0, INPUT);
        pinMode(keyRow_2, INPUT);
        pinMode(keyRow_3, INPUT);
        pinMode(keyRow_1, OUTPUT);
        digitalWrite(keyRow_1,LOW);
        delay(1);
        if (digitalRead(keyCol_0) == LOW) { keyVal = 4;}
        else if (digitalRead(keyCol_1) == LOW) { keyVal = 5;}
        else if (digitalRead(keyCol_2) == LOW) { keyVal = 6;}
        else if (digitalRead(keyCol_3) == LOW) { keyVal = 11;} // key 'B'
        else {
          pinMode(keyRow_0, INPUT);
          pinMode(keyRow_1, INPUT);
          pinMode(keyRow_3, INPUT);
          pinMode(keyRow_2, OUTPUT);
          digitalWrite(keyRow_2,LOW);
          delay(1);
          if (digitalRead(keyCol_0) == LOW) { keyVal = 7;}
          else if (digitalRead(keyCol_1) == LOW) { keyVal = 8;}
          else if (digitalRead(keyCol_2) == LOW) { keyVal = 9;}
          else if (digitalRead(keyCol_3) == LOW) { keyVal = 12;} // key 'C'
          else {
            pinMode(keyRow_0, INPUT);
            pinMode(keyRow_1, INPUT);
            pinMode(keyRow_2, INPUT);
            pinMode(keyRow_3, OUTPUT);
            digitalWrite(keyRow_3,LOW);
            delay(1);
            if (digitalRead(keyCol_0) == LOW) { keyVal = 14;} // key '*'
            else if (digitalRead(keyCol_1) == LOW) { keyVal = 0;}
            else if (digitalRead(keyCol_2) == LOW) { keyVal = 15;} // key '#'
            else if (digitalRead(keyCol_3) == LOW) { keyVal = 13;} // key 'D'
          }
        }
      }
      keypadLOW(); // set all rows LOW again
      keyState = 3; keyNew = keyVal;
      if (serialON == 1) {Serial.print("Keypad = "); Serial.println(keyVal);}
      if (serialON == 1) {Serial.print("KeypadOld = "); Serial.println(keyValOld);}
      if (ledOnCnt <1) {writeLED(LOW,LOW,LOW);} // turn ON LED's
      break;
    case 3:
      // constantly checks for switch release
      if (digitalRead(keyCol_0) == LOW || digitalRead(keyCol_1) == LOW || digitalRead(keyCol_2) == LOW || digitalRead(keyCol_3) == LOW) {
        // do nothing whilst a key is held down
        if (zMode == 1) {
          if (ledOnCnt <1) {writeLED(HIGH,HIGH,HIGH);} // turn OFF LED's
          keyState = 0;
        }
      }
      else {
        keyState = 4;
        keyDebounce = 10;  // allow 100mS debounce
      }
       break;
    case 4:
      // key release occured so wait for end of bounce
      if (digitalRead(keyCol_0) == LOW || digitalRead(keyCol_1) == LOW || digitalRead(keyCol_2) == LOW || digitalRead(keyCol_3) == LOW) {
          keyState = 3;
      }
      else {
        // still low so debounce timeout
        --keyDebounce;
        if (keyDebounce <= 0) {
          // debounce timed out so key released
          keyState = 0; keyDown = 0;
          if (ledOnCnt <1) {writeLED(HIGH,HIGH,HIGH);} // turn OFF LED's
          if (zMode == 0) {
            // after key always restart demo mode even if not changed
            demoStateOld = demoState;
            demoState = keyVal;
            demoTask = 0;  // re initialise demo task
            demoCnt = 0;  // clear counter
            demoInc = 1;  // set default increment
            demoPnt = 0;  // clear pointer
            setDitherOFF();
         }
          keyValOld = keyVal;
        }
      }
      break;
  }
}

// --------------------------------------------------------------------------------

void keypadTest() {
  // checks to see if a keypad key has been pressed
  if (digitalRead(keyCol_0) == LOW || digitalRead(keyCol_1) == LOW || digitalRead(keyCol_2) == LOW || digitalRead(keyCol_3) == LOW) {
    // one or more keys have been pressed
    keyDown = 1;  // flag key pressed
  }
  else {keyDown = 0;} // no key pressed
}

// --------------------------------------------------------------------------------

void ledBlinkWrite () {
  // blink LED's based on their values 0 - 255
  ledBlinkOn++; if (ledBlinkOn > 20) {ledBlinkOn = 1;}
  if (ledBlinkOn <10) {
    digitalWrite(ledRedPin, ledBlinkLft);
    digitalWrite(ledYelPin, ledBlinkCtr);
    digitalWrite(ledGrnPin, ledBlinkRht);
  }
  else {
    digitalWrite(ledRedPin, HIGH);
    digitalWrite(ledYelPin, HIGH);
    digitalWrite(ledGrnPin, HIGH);
  }
}

// --------------------------------------------------------------------------------

void loopSA() {
  // comes here every 100ms from main loop
  Serial.print("periodLftTgt=");
  Serial.print(periodLftTgt);
  Serial.print("  periodLftAv=");
  Serial.print(periodLftAv);
  Serial.print("  periodRhtTgt=");
  Serial.print(periodRhtTgt);
  Serial.print("  periodRhtAv=");
  Serial.println(periodRhtAv);
}

// --------------------------------------------------------------------------------

void loopSR() {
  // comes here every 100ms from main loop
  speedCntLftV = speedCntLft; speedCntRhtV = speedCntRht;
  speedCntLft = 0; speedCntRht = 0; // reset speed counters - overflow!
  if (serialON == 1) {
    Serial.print("Echo-L: ");
    Serial.print(durationLft);
    Serial.print(", Echo-C: ");
    Serial.print(durationCnt);
    Serial.print(", Echo-R: ");
    Serial.print(durationRht);
    Serial.print(", PWMA: ");
    Serial.print(speedPWMA);
    Serial.print(", PWMB: ");
    Serial.print(speedPWMB);
    Serial.print(", SpCntL: ");
    Serial.print(speedCntLftV);
    Serial.print(", SpCntR: ");
    Serial.print(speedCntRhtV);
    Serial.print(", Vin:  ");
    Serial.print(rawVin);
    Serial.println("");
  }
}

// --------------------------------------------------------------------------------

void readVin() {
  // read the analogue voltage on Pin 0
  rawVin = analogRead(pinVin);
}

// --------------------------------------------------------------------------------

void runPOST() {
  // LED light sequence on power-up.
  // Print regardless of SerialON setting.
  Serial.begin(9600); Serial.println(""); Serial.println("POST Started... ");
  int zDel = 100;
  for (int zL = 1; zL <=3; zL++) {
    writeLED(LOW, HIGH, HIGH); delay(zDel); Serial.print("1 ");
    writeLED(HIGH, LOW, HIGH); delay(zDel); Serial.print("2 ");
    writeLED(HIGH, HIGH, LOW); delay(zDel); Serial.println("3 ");
    writeLED(HIGH, HIGH, HIGH); delay(zDel);
  }
  Serial.println("POST Complete!");
  lastMicros = micros();  // resync loop timer
}

// --------------------------------------------------------------------------------

void setBrake() {
  // Set the InX pins for brake mode. 
  digitalWrite(pinIn1, LOW);
  digitalWrite(pinIn2, LOW);
  digitalWrite(pinIn3, LOW);
  digitalWrite(pinIn4, LOW);
}

// --------------------------------------------------------------------------------

void setDefaults() {
  // set default values of all variables
  anyAVal = 0; // any temp value
  anyBVal = 0; // any temp value
  autoPWMA = speedMinF; // baseline speed set for autonomous mode
  autoPWMB = speedMinF; // baseline speed set for autonomous mode
  autoSTOPPED = 0; // counts stationary conditions to trigger 'move' response
  clrCmdArray(); // clear array elements and pointer
  cmdCtrlVal = 0; // any value used in commands
  demoCnt = 0;    // a counter used by any demo
  demoCnt1 = 0;    // a counter used by any demo
  demoCntMax = 0;  // any max value
  demoFlag = 0;  // a flag value used by any demo
  demoInc = 1;  // a increment used by any demo
  demoPnt = 0;    // pointer used by any demo
  demoPntMin = 0;  // minimum pointer value in any demo
  demoState = 0;  // main state machine for demo functions
  demoStateOld = 0;  // previous task when keypad pressed
  demoTask = 0;  // current demo task
  demoWait = 0; // count down wait used in any demo
  distanceCntLft = 0; // left-hand distance count
  distanceCntRht = 0; // left-hand distance count
  ditherCnt = 2;  // sets the time period for dither = 2 * ditherCnt
  ditherLmt = 10;  // threshold at which dither is switched off
  ditherON = 0;  // 0 = OFF, 1 - 4 tasks
  ditherPWMA = 0;  // dither amplitude applied to PWMA
  ditherPWMB = 0;  // dither amplitude applied to PWMB
  ditherSpd = 2;  // value which sets the dither period
  durationCnt = 0; // measured time on centre sensor
  durationLft = 0; // measured time on leftt sensor
  durationRht = 0; // measured time on right sensor
  echoMax = 8000;  // max pulse timeout 8 ms = 8000 x 58 = 138 cm
  echoTrig = echoTrig1;  // echo pulse duration trigger point. 1 of 10 values
  irInc = 8;  // increment when using IR joypad
  irIncSpCtrl = 1000;  // speed control increment when using IR joypad
  interval = 10000; // main loop interval in microseconds
  intLftMicros = 0; // left-hand interupt period 
  intLftPrev = 0; // previous left-hand interupt period 
  intRhtMicros = 0; // right-hand interupt period 
  intRhtPrev = 0; // previous right-hand interupt period 
  keyDebounce = 0;
  keyDown = 0;  // = 1 if down, = 0 if up
  keyNew = -1; //most recent key pressed
  keyState = 0;  // state mode for debounce, etc
  keyVal = 0;  // most recent key pressed
  keyValOld = 0;  // previous key pressed
  lastMicros = micros(); // previous micros() reading
  ledBlinkCtr = 0;  // 0 - 255 blink centre value
  ledBlinkLft = 0;  // 0 - 255 blink left value
  ledBlinkOn = 0; // > 0 then blink LED's based on ledBlinkLft,...
  ledBlinkRht = 0;  // 0 - 255 blink right value
  ledOnCnt = 0;  // if > 1 then turn LED's ON every main loop cycle
  loopCnt = 0; // 1-10 counter used for 1 second service routine
  rawVin = 0; // Vin value read from analogue pin 0 - 1023
  periodDef = 70000; // period assigned to stop/default condition
  periodLft = periodDef; // left-hand interupt period
  periodLftAv = periodDef; // left-hand average interupt period
  periodLftLast = periodDef; // anti-lock period check
  periodLftTgt = 0; // left-hand period target for speed control
  periodMax = 35000; // longest period/slowest speed
  periodMin = 6000; // shortest period/fastest speed
  periodRht = periodDef; // right-hand interupt period
  periodRhtAv = periodDef; // right-hand average interupt period
  periodRhtLast = periodDef; // anti-lock period check
  periodRhtTgt = 0; // right-hand period target for speed control
//  serialON = 0;  // set to 1 to enable print responses
  spdCtrlCnt = 0; // period for speed control in 10ms steps
  spdCtrlON = 0; // speed control; 0 = OFF, 1 = ON
  spdCtrlPrd = 5; // perdiod value for speed control counter in 10ms
  spdLftAbs = 0; // anti-lock counter
  spdLftDir = 0; // cpeed control channel direction, 1 = fwd, -1 = rev
  spdLckInc = 5; // increment used in abs speed lock
  spdLckTrg = 3; // trigger count for speed lock drive
  spdRhtAbs = 0; // anti-lock counter
  spdRhtDir = 0; // cpeed control channel direction, 1 = fwd, -1 = rev
  speedCntLft = 0; // left-hand speed count
  speedCntRht = 0; // left-hand speed count
  speedInc = 5; // manual speed increment for PWM
  speedMaxF = 255;  // max forward speed value
  speedMinF = 40;  // min forward value
  speedMaxR = -speedMaxF;  // max reverse speed value
  speedMinR = -speedMinF;  // min reverse speed value
  speedPWMA = 0;  // A channel PWM
  speedPWMB = 0;  // B channel PWM
  steerBiasA = 0; // a decaying bias applied to channel A
  steerBiasIncA = 0; // decaying bias increment applied to channel A
  steerBiasB = 0; // a decaying bias applied to channel B
  steerBiasIncB = 0; // decaying bias increment applied to channel B
  steerFlag = false; // set true by IR in drive mode to return steering fwd
  //
  // Calculate variables
  ditherLmt = (speedMaxF - speedMinF) / 8;
}

// --------------------------------------------------------------------------------

void setDitherOFF() {
  // turns OFF dither mode and resets settings
  ditherCnt = 0;
  ditherON = 0; // turns it OFF
  ditherPWMA = 0;   ditherPWMB = 0;
  ditherSpd = 2;
}


// --------------------------------------------------------------------------------

void setDitherON() {
  // turns ON dither mode with initial settings
  if (spdCtrlON > 0) {setSpeedOFF();}
  setDitherOFF();
  ditherON = 1; // turns it ON
}

// --------------------------------------------------------------------------------

void setDriveSpeeds(int zL, int zR) {
  // set the speed and direction of both LH and RH drives
  if (zL > 0) {
    // set LH forward motion
    digitalWrite(pinIn1, LOW);
    digitalWrite(pinIn2, HIGH);
  }
  else if (zL <0) {
    // set LH reverse motion
    zL = -zL;
    digitalWrite(pinIn1, HIGH);
    digitalWrite(pinIn2, LOW);
  }  
  if (zR > 0) {
    // set LH forward motion
    digitalWrite(pinIn3, LOW);
    digitalWrite(pinIn4, HIGH);
  }
  else if (zR < 0) {
    // set LH reverse motion
    zR = -zR;
    digitalWrite(pinIn3, HIGH);
    digitalWrite(pinIn4, LOW);
  }  
  analogWrite(pinEnA, zL);
  analogWrite(pinEnB, zR);
}

// --------------------------------------------------------------------------------

void setForward() {
  // Set the InX pins for forward motion. 
  digitalWrite(pinIn1, LOW);
  digitalWrite(pinIn2, HIGH);
  digitalWrite(pinIn3, LOW);
  digitalWrite(pinIn4, HIGH);
}

// --------------------------------------------------------------------------------

void setLedBlinkOn() {
  // turn ON blinking LED in sensor mode
  if (ledBlinkOn < 1) {ledBlinkOn = 1;}
}

// --------------------------------------------------------------------------------

void setSpeedIntOFF() {
  // disables interrupts for speed/distance measurement
  detachInterrupt(0);
  detachInterrupt(1);
}

// --------------------------------------------------------------------------------

void setSpeedIntON() {
  // enables interrupts for speed measurement
  attachInterrupt(0, intSpeedCntRht, CHANGE); // RH interrupt
  attachInterrupt(1, intSpeedCntLft, CHANGE); // LH interrupt
  pinMode(intPin0, INPUT_PULLUP);
  pinMode(intPin1, INPUT_PULLUP);
}

// --------------------------------------------------------------------------------

void setReverse() {
  // Set the InX pins for reverse motion. 
  digitalWrite(pinIn1, HIGH);
  digitalWrite(pinIn2, LOW);
  digitalWrite(pinIn3, HIGH);
  digitalWrite(pinIn4, LOW);
}

// --------------------------------------------------------------------------------

void setSpeedOFF() {
  // turn OFF speed control
  spdCtrlON = 0;
}

// --------------------------------------------------------------------------------

void setSpeedON() {
  // turn ON speed control
  if (ditherON > 0) {setDitherOFF();}
  intLftMicros = micros(); intLftPrev = intLftMicros; speedCntLft=0;
  intRhtMicros = intLftMicros; intRhtPrev = intRhtMicros; speedCntRht=0;
  periodLftTgt = periodLftAv; // use current speeds as default
  periodRhtTgt = periodRhtAv;
  spdLftDir = 0; spdRhtDir = 0;
  if (speedPWMA > 0) {spdLftDir = 1;}
  else if (speedPWMA < 0) {spdLftDir = -1;}
  if (speedPWMB > 0) {spdRhtDir = 1;}
  else if (speedPWMB < 0) {spdRhtDir = -1;}
  spdCtrlON = 1;
}

// --------------------------------------------------------------------------------

void setTurnLeft() {
  // Set the InX pins for left turn motion. 
  digitalWrite(pinIn1, LOW);
  digitalWrite(pinIn2, HIGH);
  digitalWrite(pinIn3, HIGH);
  digitalWrite(pinIn4, LOW);
}

// --------------------------------------------------------------------------------

void setTurnRight() {
  // Set the InX pins for right turn motion. 
  digitalWrite(pinIn1, HIGH);
  digitalWrite(pinIn2, LOW);
  digitalWrite(pinIn3, LOW);
  digitalWrite(pinIn4, HIGH);
}

// --------------------------------------------------------------------------------

void softRESET() {
  // identical to a power-up reset
  setSpeedIntOFF(); delay(50); // prevent crash due to interrupts
  setup();
  waitKeyDown();
  waitIrRx();
}

// --------------------------------------------------------------------------------

void spdCtrlAdj() {
  // adjust speed every 100ms  
}

// --------------------------------------------------------------------------------

void translateIR() {
  // takes action based on IR code received
  // describing KEYES Remote IR codes
  // uses latest Ken Shirriff library v2.01
  switch(results.value) {
//    case 0x84C: Serial.println(" PWR"); break;
//    case 0x4C: Serial.println(" PWR"); break;
    case 0x84C: keyVal = 15; break;
    case 0x4C: keyVal = 15; break;
//    case 0x86F: Serial.println(" GUIDE"); break;
//    case 0x6F: Serial.println(" GUIDE"); break;
//    case 0x870: Serial.println(" MENU"); break;
//    case 0x70: Serial.println(" MENU"); break;
//    case 0x854: Serial.println(" FORWARD"); break;
//    case 0x54: Serial.println(" FORWARD"); break;
    case 0x854: keyVal = 22; break;
    case 0x54: keyVal = 22; break;
//    case 0x855: Serial.println(" LEFT"); break;
//    case 0x55: Serial.println(" LEFT"); break;
    case 0x855: keyVal = 24; break;
    case 0x55: keyVal = 24; break;
//    case 0x875: Serial.println(" -OK-"); break;
//    case 0x75: Serial.println(" -OK-"); break;
    case 0x875: keyVal = 5; break;
    case 0x75: keyVal = 5; break;
//    case 0x856: Serial.println(" RIGHT"); break;
//    case 0x56: Serial.println(" RIGHT"); break;
    case 0x856: keyVal = 26; break;
    case 0x56: keyVal = 26; break;
//    case 0x853: Serial.println(" REVERSE"); break;
//    case 0x53: Serial.println(" REVERSE"); break;
    case 0x853: keyVal = 28; break;
    case 0x53: keyVal = 28; break;
//    case 0x841: Serial.println(" 1"); break;
//    case 0x41: Serial.println(" 1"); break;
    case 0x841: keyVal = 1; break;
    case 0x41: keyVal = 1; break;
//    case 0x842: Serial.println(" 2"); break;
//    case 0x42: Serial.println(" 2"); break;
    case 0x842: keyVal = 2; break;
    case 0x42: keyVal = 2; break;
//    case 0x843: Serial.println(" 3"); break;
//    case 0x43: Serial.println(" 3"); break;
    case 0x843: keyVal = 3; break;
    case 0x43: keyVal = 3; break;
//    case 0x844: Serial.println(" 4"); break;
//    case 0x44: Serial.println(" 4"); break;
    case 0x844: keyVal = 4; break;
    case 0x44: keyVal = 4; break;
//    case 0x845: Serial.println(" 5"); break;
//    case 0x45: Serial.println(" 5"); break;
    case 0x845: keyVal = 5; break;
    case 0x45: keyVal = 5; break;
//    case 0x846: Serial.println(" 6"); break;
//    case 0x46: Serial.println(" 6"); break;
    case 0x846: keyVal = 6; break;
    case 0x46: keyVal = 6; break;
//    case 0x847: Serial.println(" 7"); break;
//    case 0x47: Serial.println(" 7"); break;
    case 0x847: keyVal = 7; break;
    case 0x47: keyVal = 7; break;
//    case 0x848: Serial.println(" 8"); break;
//    case 0x48: Serial.println(" 8"); break;
    case 0x848: keyVal = 8; break;
    case 0x48: keyVal = 8; break;
//    case 0x849: Serial.println(" 9"); break;
//    case 0x49: Serial.println(" 9"); break;
    case 0x849: keyVal = 9; break;
    case 0x49: keyVal = 9; break;
//    case 0x857: Serial.println(" -/--"); break;
//    case 0x57: Serial.println(" -/--"); break;
    case 0x857: keyVal = 14; break; // * key
    case 0x57: keyVal = 14; break;
//    case 0x840: Serial.println(" 0"); break;
//    case 0x40: Serial.println(" 0"); break;
    case 0x840: keyVal = 0; break;
    case 0x40: keyVal = 0; break;
//    case 0x878: Serial.println(" AV"); break;
//    case 0x78: Serial.println(" AV"); break;
    case 0x878: keyVal = 15; break; // # key
    case 0x78: keyVal = 15; break;
//    case 0x850: Serial.println(" Snd Up"); break;
//    case 0x50: Serial.println(" Snd Up"); break;
//    case 0x851: Serial.println(" Snd Dwn"); break;
//    case 0x51: Serial.println(" Snd Dwn"); break;
//    case 0x84D: Serial.println(" Mute"); break;
//    case 0x4D: Serial.println(" Mute"); break;
//    case 0x860: Serial.println(" Ch Up"); break;
//    case 0x60: Serial.println(" Ch Up"); break;
//    case 0x861: Serial.println(" Ch Dwn"); break;
//    case 0x61: Serial.println(" Ch Dwn"); break;
//    case 0x877: Serial.println(" Red"); break;
//    case 0x77: Serial.println(" Red"); break;
    case 0x877: keyVal = 10; break; // A key
    case 0x77: keyVal = 10; break;
//    case 0x876: Serial.println(" Grn"); break;
//    case 0x76: Serial.println(" Grn"); break;
    case 0x876: keyVal = 11; break; // B key
    case 0x76: keyVal = 11; break;
//    case 0x872: Serial.println(" Yel"); break;
//    case 0x72: Serial.println(" Yel"); break;
    case 0x872: keyVal = 12; break; // C key
    case 0x72: keyVal = 12; break;
//    case 0x874: Serial.println(" Blu"); break;
//    case 0x74: Serial.println(" Blu"); break;
    case 0x874: keyVal = 13; break; // D key
    case 0x74: keyVal = 13; break;
//    case 0x87C: Serial.println(" Txt 1"); break;
//    case 0x7C: Serial.println(" Txt 1"); break;
    case 0x87C: keyVal = 90; break; // Txt 1 key
    case 0x7C: keyVal = 90; break;
//    case 0x875: Serial.println(" Txt 2"); break;
//    case 0x75: Serial.println(" Txt 2"); break;
//    case 0x84B: Serial.println(" Txt 3"); break;
//    case 0x4B: Serial.println(" Txt 3"); break;
//    case 0x84E: Serial.println(" Txt 4"); break;
//    case 0x4E: Serial.println(" Txt 4"); break;
//    case 0xFFFFFFFF: Serial.println(" REPEAT"); break;  
  }
}

// --------------------------------------------------------------------------------

void testKeyDown() {
  // tests for a keydown condition and sets falg accordingly
  // keyDown = 1 if down, = 0 if up
  keyDown = 0;
  if (digitalRead(keyCol_0) == LOW || digitalRead(keyCol_1) == LOW || digitalRead(keyCol_2) == LOW || digitalRead(keyCol_3) == LOW) {
    // key is down!
    keyDown = 1;
  }
}

// --------------------------------------------------------------------------------

void waitIrRx(){
  // wait until IR receive stops
  int zCnt = 150; // 150mS debounce timer
  while (zCnt > 0) {
    if (irrecv.decode(&results)) {
      irrecv.resume();
      zCnt = 150;
    } delay(1); zCnt--;
  }
}

// --------------------------------------------------------------------------------

void waitKeyDown() {
  // wait here until all keys are released
  int zCnt = 100; // 100mS debounce timer
  while (zCnt > 0) {
    if (digitalRead(keyCol_0) == LOW || digitalRead(keyCol_1) == LOW || digitalRead(keyCol_2) == LOW || digitalRead(keyCol_3) == LOW) {
      // one or more keys have been pressed so wait
      zCnt = 100;
    } delay(1); zCnt--;
  }
}

// --------------------------------------------------------------------------------

void writeLED(int zL1, int zL2, int zL3) {
  // write HIGH/LOW values to the three LED's.
  // LOW  = ON
  // HIGH = OFF
  digitalWrite(ledGrnPin, zL3);
  digitalWrite(ledRedPin, zL1);
  digitalWrite(ledYelPin, zL2);
}
